/*
                        Compile
    Linux:
        gcc ./fwupd.c -I/usr/include/PCSC -lpcsclite -ofwupd

    For Windows
        x86_64-w64-mingw32-gcc ./fwupd.c -lwinscard -ofwupd.exe
*/

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#ifdef __MINGW32__
#include <winscard.h>
#else
#include <PCSC/wintypes.h>
#include <PCSC/winscard.h>
#endif

#define EUICC_INTERFACE_BUFSZ 264

#define APDU_FWUPD_SELECT_AID "\x01\xA4\x04\x00\x10\xA0\x65\x73\x74\x6B\x6D\x65\xFF\xFF\xFF\xFF\x66\x77\x75\x70\x64"
#define APDU_FWUPD_TRIGGER "\x01\x55\x55\x55\x00"

#define APDU_BOOT_APP "\x00\x00\x00\x00\x00"
#define APDU_ERASE_APP "\x00\x01\x00\x00\x00"
#define APDU_WRITE_IV "\x00\x02\x00\x00\x00"
#define APDU_READ_IV "\x00\x03\x00\x00\x00"
#define APDU_WRITE_BUFFER "\x00\x04\x00\x00\x00"
#define APDU_READ_BUFFER "\x00\x05\x00\x00\x00"
#define APDU_WRITE_APP "\x00\x06\x00\x00\x00"

#define FLASH_PAGE_SIZE 0x800

#define RAND_IV_SIZE 8
#define FLASH_BUFFER_WRITE_SIZE 128

static SCARDCONTEXT pcsc_ctx = 0;
static SCARDHANDLE pcsc_hCard = 0;
static LPSTR pcsc_mszReaders = NULL;

static int pcsc_ctx_open(void)
{
    int ret;
    DWORD dwReaders;

    pcsc_ctx = 0;
    pcsc_hCard = 0;
    pcsc_mszReaders = NULL;

    ret = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &pcsc_ctx);
    if (ret != SCARD_S_SUCCESS)
    {
        fprintf(stderr, "SCardEstablishContext() failed: %08X\n", ret);
        return -1;
    }

#ifdef SCARD_AUTOALLOCATE
    dwReaders = SCARD_AUTOALLOCATE;
    ret = SCardListReaders(pcsc_ctx, NULL, (LPSTR)&pcsc_mszReaders, &dwReaders);
#else
    // macOS does not support SCARD_AUTOALLOCATE, so we need to call SCardListReaders twice.
    // First call to get the size of the buffer, second call to get the actual data.
    ret = SCardListReaders(pcsc_ctx, NULL, NULL, &dwReaders);
    if (ret != SCARD_S_SUCCESS)
    {
        fprintf(stderr, "SCardListReaders() failed: %08X\n", ret);
        return -1;
    }
    pcsc_mszReaders = malloc(sizeof(char) * dwReaders);
    if (pcsc_mszReaders == NULL)
    {
        fprintf(stderr, "malloc: not enough memory\n");
        return -1;
    }
    ret = SCardListReaders(pcsc_ctx, NULL, pcsc_mszReaders, &dwReaders);
#endif
    if (ret != SCARD_S_SUCCESS)
    {
        fprintf(stderr, "SCardListReaders() failed: %08X\n", ret);
        return -1;
    }

    return 0;
}

static int pcsc_iter_reader(int (*callback)(int index, const char *reader))
{
    int ret;
    LPSTR psReader;

    psReader = pcsc_mszReaders;
    for (int i = 0, n = 0;; i++)
    {
        char *p = pcsc_mszReaders + i;
        if (*p == '\0')
        {
            ret = callback(n, psReader);
            if (ret < 0)
                return -1;
            if (ret > 0)
                return 0;
            if (*(p + 1) == '\0')
            {
                break;
            }
            psReader = p + 1;
            n++;
        }
    }
    return -1;
}

static int pcsc_open_hCard_iter(int index, const char *reader)
{
    int ret;
    int id;
    DWORD dwActiveProtocol;

    id = 0;
    if (getenv("DRIVER_IFID"))
    {
        id = atoi(getenv("DRIVER_IFID"));
    }

    if (id != index)
    {
        return 0;
    }

    ret = SCardConnect(pcsc_ctx, reader, SCARD_SHARE_EXCLUSIVE, SCARD_PROTOCOL_T0, &pcsc_hCard, &dwActiveProtocol);
    if (ret != SCARD_S_SUCCESS)
    {
        fprintf(stderr, "SCardConnect() failed: %08X\n", ret);
        return -1;
    }

    return 1;
}

static int pcsc_open_hCard(void)
{
    return pcsc_iter_reader(pcsc_open_hCard_iter);
}

static void pcsc_close(void)
{
    if (pcsc_mszReaders)
    {
// macOS does not support SCARD_AUTOALLOCATE, so we need to free the buffer manually.
#ifdef SCARD_AUTOALLOCATE
        SCardFreeMemory(pcsc_ctx, pcsc_mszReaders);
#else
        // on macOS, pcsc_mszReaders is allocated by malloc()
        free(pcsc_mszReaders);
#endif
    }

    if (pcsc_hCard)
    {
        SCardDisconnect(pcsc_hCard, SCARD_UNPOWER_CARD);
    }
    if (pcsc_ctx)
    {
        SCardReleaseContext(pcsc_ctx);
    }
    pcsc_ctx = 0;
    pcsc_hCard = 0;
    pcsc_mszReaders = NULL;
}

static int pcsc_transmit_lowlevel(uint8_t *rx, uint32_t *rx_len, const uint8_t *tx, uint32_t tx_len)
{
    int ret;
    DWORD rx_len_merged;

    rx_len_merged = *rx_len;
    ret = SCardTransmit(pcsc_hCard, SCARD_PCI_T0, tx, tx_len, NULL, rx, &rx_len_merged);
    if (ret != SCARD_S_SUCCESS)
    {
        fprintf(stderr, "SCardTransmit() failed: %08X\n", ret);
        return -1;
    }

    *rx_len = rx_len_merged;
    return 0;
}

static int card_poweron(void)
{
    if (pcsc_ctx_open() < 0)
    {
        return -1;
    }

    if (pcsc_open_hCard() < 0)
    {
        return -1;
    }

    return 0;
}

static void card_poweroff(void)
{
    pcsc_close();
}

static void send_magic(void)
{
    uint8_t pcsc_rx[EUICC_INTERFACE_BUFSZ];
    uint32_t pcsc_rx_len;

    pcsc_rx_len = sizeof(pcsc_rx);
    memset(pcsc_rx, 0, sizeof(pcsc_rx));
    pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, APDU_FWUPD_SELECT_AID, sizeof(APDU_FWUPD_SELECT_AID) - 1);

    pcsc_rx_len = sizeof(pcsc_rx);
    memset(pcsc_rx, 0, sizeof(pcsc_rx));
    pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, APDU_FWUPD_TRIGGER, sizeof(APDU_FWUPD_TRIGGER) - 1);
}

static int erase_app(void)
{
    uint8_t pcsc_rx[EUICC_INTERFACE_BUFSZ];
    uint32_t pcsc_rx_len;

    pcsc_rx_len = sizeof(pcsc_rx);
    memset(pcsc_rx, 0, sizeof(pcsc_rx));
    pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, APDU_ERASE_APP, sizeof(APDU_ERASE_APP) - 1);
    if (pcsc_rx[0] != 0x90)
    {
        return -1;
    }
    return 0;
}

static int send_sector(FILE *fwfp, uint8_t sector)
{
    uint8_t randiv[RAND_IV_SIZE];
    uint8_t flash_sector_buffer[FLASH_PAGE_SIZE];

    uint8_t pcsc_tx[EUICC_INTERFACE_BUFSZ];
    uint8_t pcsc_rx[EUICC_INTERFACE_BUFSZ];
    uint32_t pcsc_rx_len;

    if (fseek(fwfp, sector * (RAND_IV_SIZE + FLASH_PAGE_SIZE), SEEK_SET) < 0)
    {
        return -1;
    }

    if (fread(randiv, 1, RAND_IV_SIZE, fwfp) < 0)
    {
        return -1;
    }

    if (fread(flash_sector_buffer, 1, FLASH_PAGE_SIZE, fwfp) < 0)
    {
        return -1;
    }

    if (feof(fwfp))
    {
        return 1;
    }

    memcpy(pcsc_tx, APDU_WRITE_IV, sizeof(APDU_WRITE_IV) - 1);
    memcpy(pcsc_tx + sizeof(APDU_WRITE_IV) - 1, randiv, RAND_IV_SIZE);
    pcsc_tx[4] = RAND_IV_SIZE;

    pcsc_rx_len = sizeof(pcsc_rx);
    memset(pcsc_rx, 0, sizeof(pcsc_rx));
    pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, pcsc_tx, RAND_IV_SIZE + sizeof(APDU_WRITE_IV) - 1);
    if (pcsc_rx[0] != 0x90)
    {
        fprintf(stderr, "iv fill failed: %u\n", sector);
        return -1;
    }

    memcpy(pcsc_tx, APDU_READ_IV, sizeof(APDU_READ_IV) - 1);
    pcsc_tx[4] = RAND_IV_SIZE;

    pcsc_rx_len = sizeof(pcsc_rx);
    memset(pcsc_rx, 0, sizeof(pcsc_rx));
    pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, pcsc_tx, sizeof(APDU_READ_IV) - 1);
    if (memcmp(pcsc_rx, randiv, RAND_IV_SIZE) != 0)
    {
        fprintf(stderr, "iv verify failed: %u\n", sector);
        return -1;
    }

    for (int i = 0; i < FLASH_PAGE_SIZE / FLASH_BUFFER_WRITE_SIZE; i++)
    {
        memcpy(pcsc_tx, APDU_WRITE_BUFFER, sizeof(APDU_WRITE_BUFFER) - 1);
        memcpy(pcsc_tx + sizeof(APDU_WRITE_BUFFER) - 1, flash_sector_buffer + (i * FLASH_BUFFER_WRITE_SIZE), FLASH_BUFFER_WRITE_SIZE);

        pcsc_tx[2] = (i * FLASH_BUFFER_WRITE_SIZE) >> 8;
        pcsc_tx[3] = (i * FLASH_BUFFER_WRITE_SIZE) & 0xFF;
        pcsc_tx[4] = FLASH_BUFFER_WRITE_SIZE;

        pcsc_rx_len = sizeof(pcsc_rx);
        memset(pcsc_rx, 0, sizeof(pcsc_rx));
        pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, pcsc_tx, FLASH_BUFFER_WRITE_SIZE + sizeof(APDU_WRITE_BUFFER) - 1);
        if (pcsc_rx[0] != 0x90)
        {
            fprintf(stderr, "buffer fill failed: %u-%u\n", sector, i);
            return -1;
        }

        memcpy(pcsc_tx, APDU_READ_BUFFER, sizeof(APDU_READ_BUFFER) - 1);
        pcsc_tx[2] = (i * FLASH_BUFFER_WRITE_SIZE) >> 8;
        pcsc_tx[3] = (i * FLASH_BUFFER_WRITE_SIZE) & 0xFF;
        pcsc_tx[4] = FLASH_BUFFER_WRITE_SIZE;

        pcsc_rx_len = sizeof(pcsc_rx);
        memset(pcsc_rx, 0, sizeof(pcsc_rx));
        pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, pcsc_tx, sizeof(APDU_READ_BUFFER) - 1);
        if (memcmp(pcsc_rx, flash_sector_buffer + (i * FLASH_BUFFER_WRITE_SIZE), FLASH_BUFFER_WRITE_SIZE) != 0)
        {
            fprintf(stderr, "buffer verify failed: %u-%u\n", sector, i);
            i--;
            continue;
        }
    }

    return 0;
}

static int write_app(uint8_t sector)
{
    uint8_t pcsc_tx[EUICC_INTERFACE_BUFSZ];
    uint8_t pcsc_rx[EUICC_INTERFACE_BUFSZ];
    uint32_t pcsc_rx_len;

    memcpy(pcsc_tx, APDU_WRITE_APP, sizeof(APDU_WRITE_APP) - 1);
    pcsc_tx[2] = 0;
    pcsc_tx[3] = sector;
    pcsc_rx_len = sizeof(pcsc_rx);
    memset(pcsc_rx, 0, sizeof(pcsc_rx));
    pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, pcsc_tx, sizeof(APDU_WRITE_APP) - 1);
    if (pcsc_rx[0] != 0x90)
    {
        return -1;
    }
    return 0;
}

static int boot_app(void)
{
    uint8_t pcsc_rx[EUICC_INTERFACE_BUFSZ];
    uint32_t pcsc_rx_len;

    pcsc_rx_len = sizeof(pcsc_rx);
    memset(pcsc_rx, 0, sizeof(pcsc_rx));
    pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, APDU_BOOT_APP, sizeof(APDU_BOOT_APP) - 1);
    if (pcsc_rx[0] != 0x90)
    {
        return -1;
    }
    return 0;
}

int fwupd()
{
    FILE *fw;

    fw = fopen("estk.app.bin", "rb");
    if (fw == NULL)
    {
        fprintf(stderr, "fw file estk.app.bin fopen() failed\n");
        return -1;
    }

    card_poweron();
    card_poweroff();
    if (card_poweron() < 0)
    {
        fprintf(stderr, "open smartcard 1 failed\n");
        return -1;
    }

    send_magic();

    card_poweroff();
    card_poweron();
    card_poweroff();
    if (card_poweron() < 0)
    {
        fprintf(stderr, "open smartcard 2 failed\n");
        return -1;
    }

    if (erase_app() < 0)
    {
        fprintf(stderr, "app erase failed\n");
        return -1;
    }
    printf("app erase done\n");

    for (int i = 1;; i++)
    {
        int ret;
        printf("sending sector %u\n", i);
        ret = send_sector(fw, i);
        if (ret == -1)
        {
            fprintf(stderr, "sector %u send failed\n", i);
            return -1;
        }
        else if (ret == 1)
        {
            break;
        }
        printf("sector %u send done\n", i);

        if (write_app(i))
        {
            fprintf(stderr, "sector %u write failed\n", i);
        }
        printf("sector %u write done\n", i);
    }

    if (send_sector(fw, 0) < 0)
    {
        fprintf(stderr, "sector 0 send failed\n");
        return -1;
    }
    printf("sector 0 send done\n");

    if (write_app(0))
    {
        fprintf(stderr, "sector 0 write failed\n");
    }
    printf("sector 0 write done\n");

    if (boot_app() < 0)
    {
        fprintf(stderr, "boot app failed\n");
        return -1;
    }
    printf("boot app done\n");

    sleep(1);

    pcsc_close();

    return 0;
}

int main(int argc, char *argv[])
{
    int ret;

    ret = fwupd();

    printf("Press any key to continue...\n");
    getchar();

    return ret;
}
